#ifndef RESULT_H
#define RESULT_H

#include <stdbool.h>

/**
 * @file result.h
 * @brief Result and Maybe types.
 */

/**
 * @brief Result with value and status.
 *
 * Usage:
 * @code
 * typedef enum {
 *     CMD_ARGS_CORRECT = 0,
 *     CMD_ARGS_INCORRECT,
 *     CMD_ARGS_BAD_ANGLE,
 * } ArgsParseStatus;
 *
 * Result(ArgsParseResult, CmdArgs, ArgsParseStatus);
 *
 * ArgsParseResult parsed_args = do_stuff();
 *
 * if (parsed_args.status != CMD_ARGS_CORRECT) {
 *     exit(1);
 * }
 * print_args(parsed_args._);
 * @endcode
 */
#define Result(name, value_t, status_t)                                        \
    typedef struct {                                                           \
        status_t status;                                                       \
        value_t _;                                                             \
    }(name)

/**
 * @brief Maybe result.
 *
 * Maybe is a value tagged with a validness flag.
 *
 * Usage:
 * header.h
 * @code
 * Maybe(Image);
 * @endcode
 *
 * source.c
 * @code
 * MaybeImpl(Image);
 *
 * MaybeImage make_image() {
 *     // ...
 *     if (some_condition) {
 *         return SomeImage(img);
 *     }
 *     return NoneImage;
 * }
 * @endcode
 */
#define Maybe(value_t)                                                         \
    Result(Maybe##value_t, value_t, bool);                                     \
                                                                               \
    Maybe##value_t Some##value_t(value_t value);                               \
                                                                               \
    extern const Maybe##value_t None##value_t

#define MaybeImpl(value_t)                                                     \
    Maybe##value_t Some##value_t(value_t value) {                              \
        return (Maybe##value_t){true, value};                                  \
    }                                                                          \
                                                                               \
    const Maybe##value_t None##value_t = {false}

/* Result type for IO operations. */
#define IOResult(name, value_t) Result(name, value_t, IOStatus)

/* IO operations statuses. */
typedef enum {
    /* Keep calm, everything is fine. */
    IO_OK = 0,
    /* Stream opening error. */
    IO_OPEN_ERR,
    /* Stream closing error. You may lost written data probably. */
    IO_CLOSE_ERR,
    /* Cannot read from stream. */
    IO_READ_ERR,
    /* Cannot write to stream. */
    IO_WRITE_ERR,
} IOStatus;

#endif /* RESULT_H */
