#include "bmp.h"

#include "image_io.h"

#include <inttypes.h>
#include <memory.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>

typedef struct __attribute__((packed)) {
    /* Bitmap file header-> */
    struct __attribute__((packed)) {
        /* The header field used to identify the BMP and DIB file. Set to 0x42
         * 0x4D in hexadecimal, same as BM in ASCII. */
        uint16_t type;
        /* The size of the BMP file in bytes. */
        uint32_t file_size;
        /* Reserved. */
        uint32_t reserved;
        /* Starting address of the byte where the bitmap pixel array starts. */
        uint32_t pixels_offset;
    } bf;
    /* Bitmap information header (Windows BITMAPINFOHEADER). */
    struct __attribute__((packed)) {
        /* The size of this header (40 bytes). */
        uint32_t header_size;
        /* Bitmap width in pixels */
        uint32_t width;
        /* Bitmap height in pixels. */
        uint32_t height;
        /* Number of color planes (must be 1). */
        uint16_t planes;
        /* Number of bits per pixel. */
        uint16_t bit_count;
        /* Compression method. */
        uint32_t compression;
        /* Image size. This is the size of the raw bitmap data. */
        uint32_t image_size;
        /* Horizontal resolution of the image. */
        uint32_t x_pix_per_meter;
        /* Vertical resolution of the image. */
        uint32_t y_pix_per_meter;
        /* Number of colors in the color palette, or 0 to default to 2n. */
        uint32_t colors_used;
        /* Number of important colors, 0 when every color is important. */
        uint32_t colors_important;
    } bi;
} BmpHeader;

static const size_t PIXEL_SIZE = sizeof(Pixel); // Pixel size in bytes

/* From BMP specification. */
BmpHeader _HEADER = {0};
static const uint16_t BMP_FILE_TYPE = 0x4D42;            // Little-endian 'BM'
static const uint32_t BMP_RESERVED = 0;                  // Always 0
static const size_t BMP_HEADER_SIZE = sizeof(_HEADER);   // size of full header
static const size_t BI_HEADER_SIZE = sizeof(_HEADER.bi); // 40 bytes
// We support only uncompressed files
static const uint32_t BI_RGB_COMPRESSION = 0;
static const uint16_t BI_PLANES = 1;     // Always 1
static const uint32_t BI_ALL_COLORS = 0; // Enable full color palette
static const uint32_t BI_DUMMY_SIZE = 0; // Image size for not-compressed images
static const uint16_t COLOR_DEPTH = (PIXEL_SIZE * 8); // 24 bits
static const uint32_t DPI_72 = 2835;   // Default image resolution
static const uint8_t PADDING_BYTE = 0; // padding stab

#define array_pixel_at(pixels_array, row, col, row_size)                       \
    ((pixels_array) + (row) * (row_size) + (col)*PIXEL_SIZE)

static uint8_t _calc_padding(size_t row_size) {
    return (4 - row_size % 4) % 4;
}

static FromBmpStatus _validate_header(BmpHeader *header) {
    if (header->bf.type != BMP_FILE_TYPE) {
        return FROM_BMP_UNSUPPORTED_FORMAT;
    }
    if (header->bi.header_size != BI_HEADER_SIZE) {
        return FROM_BMP_UNSUPPORTED_FORMAT;
    }
    if (header->bi.compression != BI_RGB_COMPRESSION) {
        return FROM_BMP_UNSUPPORTED_COMPRESSION;
    }
    if (header->bi.bit_count != COLOR_DEPTH) {
        return FROM_BMP_UNSUPPORTED_COLOR_DEPTH;
    }
    return FROM_BMP_OK;
}

static void _set_pixel_to_image(
    uint8_t const *pixels_array,
    Image img,
    size_t row,
    size_t col,
    size_t row_size
) {
    uint8_t const *pixel_ptr = array_pixel_at(pixels_array, row, col, row_size);
    *pixel_at(img, col, row) = (Pixel){
        .b = pixel_ptr[0],
        .g = pixel_ptr[1],
        .r = pixel_ptr[2],
    };
}

ImageParserResult from_bmp(FileBuffer buffer) {
    if (buffer.size < BMP_HEADER_SIZE) {
        return (ImageParserResult){FROM_BMP_INVALID_HEADER};
    }
    BmpHeader *header = (void *)buffer.data;
    if (buffer.size < header->bf.file_size) {
        return (ImageParserResult){FROM_BMP_INVALID_PIXELS};
    }

    const FromBmpStatus header_status = _validate_header(header);
    if (header_status != FROM_BMP_OK) {
        return (ImageParserResult){header_status};
    }

    MaybeImage img = create_image(header->bi.width, header->bi.height);
    if (!img.status) {
        return (ImageParserResult){FROM_BMP_CANNOT_ALLOC_MEMORY};
    }

    const uint8_t padding = _calc_padding(img._.width * PIXEL_SIZE);
    const size_t row_size = img._.width * PIXEL_SIZE + padding;
    const uint8_t *pixels_array = buffer.data + header->bf.pixels_offset;
    for (size_t row = 0; row < img._.height; row++) {
        for (size_t col = 0; col < img._.width; col++) {
            _set_pixel_to_image(pixels_array, img._, row, col, row_size);
        }
    }
    return (ImageParserResult){FROM_BMP_OK, img._};
}

static size_t _calc_file_size(
    uint32_t width,
    uint32_t height,
    uint32_t padding
) {
    return BMP_HEADER_SIZE + (width * PIXEL_SIZE + padding) * height;
}

/* Would be inlined, so it's fine to return large struct. */
static BmpHeader *_build_header(
    uint32_t width,
    uint32_t height,
    size_t file_size,
    BmpHeader *header
) {
    *header = (BmpHeader){
        .bf =
            {
                .type = BMP_FILE_TYPE,
                .file_size = file_size,
                .reserved = BMP_RESERVED,
                .pixels_offset = BMP_HEADER_SIZE,
            },
        .bi =
            {
                .header_size = BI_HEADER_SIZE,
                .width = width,
                .height = height,
                .planes = BI_PLANES,
                .bit_count = COLOR_DEPTH,
                .image_size = BI_DUMMY_SIZE,
                .x_pix_per_meter = DPI_72,
                .y_pix_per_meter = DPI_72,
                .colors_used = BI_ALL_COLORS,
                .colors_important = BI_ALL_COLORS,
            },
    };
    return header;
}

/* Please, compiler, inline that >_< */
static void _set_pixel_to_buffer(
    uint8_t *pixels_array,
    Image img,
    size_t row,
    size_t col,
    size_t row_size
) {
    uint8_t *pixel_ptr = array_pixel_at(pixels_array, row, col, row_size);
    const Pixel pixel = *pixel_at(img, col, row);
    pixel_ptr[0] = pixel.b;
    pixel_ptr[1] = pixel.g;
    pixel_ptr[2] = pixel.r;
}

static void _fill_padding(
    uint8_t *pixels_array,
    size_t row,
    size_t row_size,
    uint8_t padding
) {
    const size_t padding_start = row * row_size + row_size - padding;
    for (size_t col = 0; col < padding; col++) {
        pixels_array[padding_start + col] = PADDING_BYTE;
    }
}

ImageBuilderResult to_bmp(Image img) {
    const uint8_t padding = _calc_padding(img.width * PIXEL_SIZE);
    const size_t file_size = _calc_file_size(img.width, img.height, padding);
    const MaybeFileBuffer buffer = create_buffer(file_size);
    if (!buffer.status) {
        return (ImageBuilderResult){TO_BMP_CANNOT_ALLOC_MEMORY};
    }
    BmpHeader *header = _build_header(
        img.width, img.height, file_size, (void *)buffer._.data
    );

    const size_t row_size = img.width * PIXEL_SIZE + padding;
    uint8_t *pixels_array = buffer._.data + header->bf.pixels_offset;
    for (size_t row = 0; row < img.height; row++) {
        for (size_t col = 0; col < img.width; col++) {
            _set_pixel_to_buffer(pixels_array, img, row, col, row_size);
        }
        _fill_padding(pixels_array, row, row_size, padding);
    }
    return (ImageBuilderResult){TO_BMP_OK, buffer._};
}
